@@ -9,38 +9,41 @@ raven.base
 from __future__ import absolute_import
 
 import base64
-import datetime
-import hashlib
+import zlib
 import logging
 import os
 import sys
 import time
-import urllib2
 import uuid
 import warnings
 
+from datetime import datetime
+
 import raven
 from raven.conf import defaults
 from raven.context import Context
-from raven.utils import json, get_versions, get_auth_header
-from raven.utils.encoding import to_string
+from raven.utils import six, json, get_versions, get_auth_header, merge_dicts
+from raven.utils.encoding import to_unicode
 from raven.utils.serializer import transform
-from raven.utils.stacks import get_stack_info, iter_stack_frames, \
-  get_culprit
+from raven.utils.stacks import get_stack_info, iter_stack_frames, get_culprit
 from raven.utils.urlparse import urlparse
+from raven.utils.compat import HTTPError
 from raven.transport.registry import TransportRegistry, default_transports
 
 __all__ = ('Client',)
 
 PLATFORM_NAME = 'python'
 
+# singleton for the client
+Raven = None
+
 
 class ModuleProxyCache(dict):
     def __missing__(self, key):
         module, class_name = key.rsplit('.', 1)
 
-        handler = getattr(__import__(module, {},
-                {}, [class_name], -1), class_name)
+        handler = getattr(__import__(
+            module, {}, {}, [class_name]), class_name)
 
         self[key] = handler
 
@@ -98,15 +101,6 @@ class Client(object):
     >>> # Specify a DSN explicitly
     >>> client = Client(dsn='https://public_key:secret_key@sentry.local/project_id')
 
-    >>> # Configure the client manually
-    >>> client = Client(
-    >>>     servers=['http://sentry.local/api/store/'],
-    >>>     include_paths=['my.package'],
-    >>>     project='project_id',
-    >>>     public_key='public_key',
-    >>>     secret_key='secret_key',
-    >>> )
-
     >>> # Record an exception
     >>> try:
     >>>     1/0
@@ -115,18 +109,22 @@ class Client(object):
     >>>     print "Exception caught; reference is %s" % ident
     """
     logger = logging.getLogger('raven')
-    protocol_version = '2.0'
+    protocol_version = '4'
 
     _registry = TransportRegistry(transports=default_transports)
 
     def __init__(self, dsn=None, **options):
+        global Raven
+
         o = options
 
+        self.configure_logging()
+
         # configure loggers first
         cls = self.__class__
         self.state = ClientState()
-        self.logger = logging.getLogger('%s.%s' % (cls.__module__,
-            cls.__name__))
+        self.logger = logging.getLogger(
+            '%s.%s' % (cls.__module__, cls.__name__))
         self.error_logger = logging.getLogger('sentry.errors')
 
         if dsn is None and os.environ.get('SENTRY_DSN'):
@@ -137,34 +135,41 @@ class Client(object):
         if dsn:
             # TODO: should we validate other options werent sent?
             urlparts = urlparse(dsn)
-            msg = "Configuring Raven for host: %s://%s:%s" % (urlparts.scheme,
-                    urlparts.netloc, urlparts.path)
-            self.logger.debug(msg)
+            self.logger.debug(
+                "Configuring Raven for host: %s://%s:%s" % (urlparts.scheme,
+                urlparts.netloc, urlparts.path))
             dsn_config = raven.load(dsn, transport_registry=self._registry)
             servers = dsn_config['SENTRY_SERVERS']
             project = dsn_config['SENTRY_PROJECT']
             public_key = dsn_config['SENTRY_PUBLIC_KEY']
             secret_key = dsn_config['SENTRY_SECRET_KEY']
+            transport_options = dsn_config.get('SENTRY_TRANSPORT_OPTIONS', {})
         else:
+            if o.get('servers'):
+                warnings.warn('Manually configured connections are deprecated. Switch to a DSN.', DeprecationWarning)
             servers = o.get('servers')
             project = o.get('project')
             public_key = o.get('public_key')
             secret_key = o.get('secret_key')
+            transport_options = {}
 
         self.servers = servers
         self.public_key = public_key
         self.secret_key = secret_key
         self.project = project or defaults.PROJECT
+        self.transport_options = transport_options
 
         self.include_paths = set(o.get('include_paths') or [])
         self.exclude_paths = set(o.get('exclude_paths') or [])
-        self.name = unicode(o.get('name') or defaults.NAME)
-        self.auto_log_stacks = bool(o.get('auto_log_stacks') or
-                defaults.AUTO_LOG_STACKS)
-        self.string_max_length = int(o.get('string_max_length') or
-                defaults.MAX_LENGTH_STRING)
-        self.list_max_length = int(o.get('list_max_length') or defaults.MAX_LENGTH_LIST)
+        self.name = six.text_type(o.get('name') or defaults.NAME)
+        self.auto_log_stacks = bool(
+            o.get('auto_log_stacks') or defaults.AUTO_LOG_STACKS)
+        self.string_max_length = int(
+            o.get('string_max_length') or defaults.MAX_LENGTH_STRING)
+        self.list_max_length = int(
+            o.get('list_max_length') or defaults.MAX_LENGTH_LIST)
         self.site = o.get('site', defaults.SITE)
+        self.include_versions = o.get('include_versions', True)
         self.processors = o.get('processors')
         if self.processors is None:
             self.processors = defaults.PROCESSORS
@@ -173,22 +178,40 @@ class Client(object):
         if context is None:
             context = {'sys.argv': sys.argv[:]}
         self.extra = context
+        self.tags = o.get('tags') or {}
 
         self.module_cache = ModuleProxyCache()
 
         # servers may be set to a NoneType (for Django)
         if not self.is_enabled():
-            self.logger.info('Raven is not configured (disabled). Please see documentation for more information.')
+            self.logger.info(
+                'Raven is not configured (logging is disabled). Please see the'
+                ' documentation for more information.')
+
+        if Raven is None:
+            Raven = self
+
+        self._context = Context()
 
     @classmethod
     def register_scheme(cls, scheme, transport_class):
         cls._registry.register_scheme(scheme, transport_class)
 
+    def configure_logging(self):
+        for name in ('raven', 'sentry'):
+            logger = logging.getLogger(name)
+            if logger.handlers:
+                continue
+            logger.addHandler(logging.StreamHandler())
+            logger.setLevel(logging.INFO)
+
     def get_processors(self):
         for processor in self.processors:
             yield self.module_cache[processor](self)
 
     def get_module_versions(self):
+        if not self.include_versions:
+            return {}
         return get_versions(self.include_paths)
 
     def get_ident(self, result):
@@ -203,9 +226,34 @@ class Client(object):
     def get_handler(self, name):
         return self.module_cache[name](self)
 
+    def _get_public_dsn(self):
+        url = urlparse(self.servers[0])
+        netloc = url.hostname
+        if url.port:
+            netloc += ':%s' % url.port
+        path = url.path.replace('api/%s/store/' % (self.project,), self.project)
+        return '//%s@%s%s' % (self.public_key, netloc, path)
+
+    def get_public_dsn(self, scheme=None):
+        """
+        Returns a public DSN which is consumable by raven-js
+
+        >>> # Return scheme-less DSN
+        >>> print client.get_public_dsn()
+
+        >>> # Specify a scheme to use (http or https)
+        >>> print client.get_public_dsn('https')
+        """
+        if not self.is_enabled():
+            return
+        url = self._get_public_dsn()
+        if not scheme:
+            return url
+        return '%s:%s' % (scheme, url)
+
     def build_msg(self, event_type, data=None, date=None,
-            time_spent=None, extra=None, stack=None, public_key=None,
-            tags=None, **kwargs):
+                  time_spent=None, extra=None, stack=None, public_key=None,
+                  tags=None, **kwargs):
         """
         Captures, processes and serializes an event into a dict object
 
@@ -216,12 +264,12 @@ class Client(object):
         # create ID client-side so that it can be passed to application
         event_id = uuid.uuid4().hex
 
-        if data is None:
-            data = {}
-        if extra is None:
-            extra = {}
-        if not date:
-            date = datetime.datetime.utcnow()
+        data = merge_dicts(self.context.data, data)
+
+        data.setdefault('tags', {})
+        data.setdefault('extra', {})
+        data.setdefault('level', logging.ERROR)
+
         if stack is None:
             stack = self.auto_log_stacks
 
@@ -237,7 +285,7 @@ class Client(object):
         if data.get('culprit'):
             culprit = data['culprit']
 
-        for k, v in result.iteritems():
+        for k, v in six.iteritems(result):
             if k not in data:
                 data[k] = v
 
@@ -251,8 +299,7 @@ class Client(object):
             data.update({
                 'sentry.interfaces.Stacktrace': {
                     'frames': get_stack_info(frames,
-                        list_max_length=self.list_max_length,
-                        string_max_length=self.string_max_length)
+                        transformer=self.transform)
                 },
             })
 
@@ -262,24 +309,24 @@ class Client(object):
                     if frame.get('in_app') is not None:
                         continue
 
-                    if frame.get('module') and frame.get('function'):
-                        func_name = '%s.%s' % (frame['module'], frame['module'])
-                    elif frame.get('module'):
-                        func_name = frame['module']
-                    elif frame.get('function'):
-                        func_name = frame['function']
+                    path = frame.get('module')
+                    if not path:
+                        continue
+
+                    if path.startswith('raven.'):
+                        frame['in_app'] = False
                     else:
-                        func_name = ''
+                        frame['in_app'] = (
+                            any(path.startswith(x) for x in self.include_paths)
+                            and not
+                            any(path.startswith(x) for x in self.exclude_paths)
+                        )
 
-                    func_name = '%s.%s' % (frame.get('module') or '<unknown>',
-                        frame.get('function') or '<unknown>')
-                    frame['in_app'] = (any(func_name.startswith(x)
-                        for x in self.include_paths)
-                            and not any(func_name.startswith(x)
-                                for x in self.exclude_paths))
-
-            if not culprit:
+        if not culprit:
+            if 'sentry.interfaces.Stacktrace' in data:
                 culprit = get_culprit(data['sentry.interfaces.Stacktrace']['frames'])
+            elif data.get('sentry.interfaces.Exception', {}).get('stacktrace'):
+                culprit = get_culprit(data['sentry.interfaces.Exception']['stacktrace']['frames'])
 
         if not data.get('level'):
             data['level'] = kwargs.get('level') or logging.ERROR
@@ -290,36 +337,17 @@ class Client(object):
         if not data.get('modules'):
             data['modules'] = self.get_module_versions()
 
-        data['tags'] = tags or {}
-        data.setdefault('extra', {})
-        data.setdefault('level', logging.ERROR)
+        data['tags'] = merge_dicts(self.tags, data['tags'], tags)
+        data['extra'] = merge_dicts(self.extra, data['extra'], extra)
 
-        # Add extra context
-        if self.extra:
-            for k, v in self.extra.iteritems():
-                data['extra'].setdefault(k, v)
-
-        for k, v in extra.iteritems():
-            data['extra'][k] = v
+        # Legacy support for site attribute
+        site = data.pop('site', None) or self.site
+        if site:
+            data['tags'].setdefault('site', site)
 
         if culprit:
             data['culprit'] = culprit
 
-        if not data.get('checksum'):
-            checksum_bits = handler.get_hash(data)
-        else:
-            checksum_bits = data['checksum']
-
-        if isinstance(checksum_bits, (list, tuple)):
-            checksum = hashlib.md5()
-            for bit in checksum_bits:
-                checksum.update(to_string(bit))
-            checksum = checksum.hexdigest()
-        else:
-            checksum = checksum_bits
-
-        data['checksum'] = checksum
-
         # Run the data through processors
         for processor in self.get_processors():
             data.update(processor.process(data))
@@ -327,42 +355,81 @@ class Client(object):
         if 'message' not in data:
             data['message'] = handler.to_string(data)
 
-        data.setdefault('project', self.project)
+        # tags should only be key=>u'value'
+        for key, value in six.iteritems(data['tags']):
+            data['tags'][key] = to_unicode(value)
 
-        # Legacy support for site attribute
-        site = data.pop('site', None) or self.site
-        if site:
-            data['tags'].setdefault('site', site)
-
-        # Make sure all data is coerced
-        data = self.transform(data)
+        # extra data can be any arbitrary value
+        for k, v in six.iteritems(data['extra']):
+            data['extra'][k] = self.transform(v)
 
         # It's important date is added **after** we serialize
-        data.update({
-            'timestamp': date,
-            'time_spent': time_spent,
-            'event_id': event_id,
-            'platform': PLATFORM_NAME,
-        })
+        data.setdefault('project', self.project)
+        data.setdefault('timestamp', date or datetime.utcnow())
+        data.setdefault('time_spent', time_spent)
+        data.setdefault('event_id', event_id)
+        data.setdefault('platform', PLATFORM_NAME)
 
         return data
 
     def transform(self, data):
-        return transform(data, list_max_length=self.list_max_length,
+        return transform(
+            data, list_max_length=self.list_max_length,
             string_max_length=self.string_max_length)
 
-    def context(self, **kwargs):
+    @property
+    def context(self):
         """
-        Create default context around a block of code for exception management.
+        Updates this clients thread-local context for future events.
 
-        >>> with client.context(tags={'key': 'value'}) as raven:
-        >>>     # use the context manager's client reference
-        >>>     raven.captureMessage('hello!')
-        >>>
-        >>>     # uncaught exceptions also contain the context
-        >>>     1 / 0
+        >>> def view_handler(view_func, *args, **kwargs):
+        >>>     client.context.merge(tags={'key': 'value'})
+        >>>     try:
+        >>>         return view_func(*args, **kwargs)
+        >>>     finally:
+        >>>         client.context.clear()
         """
-        return Context(self, **kwargs)
+        return self._context
+
+    def user_context(self, data):
+        """
+        Update the user context for future events.
+
+        >>> client.user_context({'email': 'foo@example.com'})
+        """
+        return self.context.merge({
+            'sentry.interfaces.User': data,
+        })
+
+    def http_context(self, data, **kwargs):
+        """
+        Update the http context for future events.
+
+        >>> client.http_context({'url': 'http://example.com'})
+        """
+        return self.context.merge({
+            'sentry.interfaces.Http': data,
+        })
+
+    def extra_context(self, data, **kwargs):
+        """
+        Update the extra context for future events.
+
+        >>> client.extra_context({'foo': 'bar'})
+        """
+        return self.context.merge({
+            'extra': data,
+        })
+
+    def tags_context(self, data, **kwargs):
+        """
+        Update the tags context for future events.
+
+        >>> client.tags_context({'version': '1.0'})
+        """
+        return self.context.merge({
+            'tags': data,
+        })
 
     def capture(self, event_type, data=None, date=None, time_spent=None,
                 extra=None, stack=None, tags=None, **kwargs):
@@ -408,7 +475,8 @@ class Client(object):
                            interfaces. Any key which contains a '.' will be
                            assumed to be a data interface.
         :param date: the datetime of this event
-        :param time_spent: a float value representing the duration of the event
+        :param time_spent: a integer value representing the duration of the
+                           event (in milliseconds)
         :param event_id: a 32-length unique string identifying this event
         :param extra: a dictionary of additional standard metadata
         :param culprit: a string representing the cause of this event
@@ -419,18 +487,13 @@ class Client(object):
         if not self.is_enabled():
             return
 
-        data = self.build_msg(event_type, data, date, time_spent,
-                extra, stack, tags=tags, **kwargs)
+        data = self.build_msg(
+            event_type, data, date, time_spent, extra, stack, tags=tags,
+            **kwargs)
 
         self.send(**data)
 
-        return (data['event_id'], data['checksum'])
-
-    def _send_remote(self, url, data, headers={}):
-        parsed = urlparse(url)
-
-        transport = self._registry.get_transport(parsed)
-        return transport.send(data, headers)
+        return (data.get('event_id'),)
 
     def _get_log_message(self, data):
         # decode message so we can show the actual event
@@ -449,29 +512,48 @@ class Client(object):
         """
         return bool(self.servers)
 
+    def _successful_send(self):
+        self.state.set_success()
+
+    def _failed_send(self, e, url, data):
+        if isinstance(e, HTTPError):
+            body = e.read()
+            self.error_logger.error(
+                'Unable to reach Sentry log server: %s (url: %s, body: %s)',
+                e, url, body, exc_info=True,
+                extra={'data': {'body': body[:200], 'remote_url': url}})
+        else:
+            self.error_logger.error(
+                'Unable to reach Sentry log server: %s (url: %s)', e, url,
+                exc_info=True, extra={'data': {'remote_url': url}})
+
+        message = self._get_log_message(data)
+        self.error_logger.error('Failed to submit message: %r', message)
+        self.state.set_fail()
+
     def send_remote(self, url, data, headers={}):
         if not self.state.should_try():
             message = self._get_log_message(data)
             self.error_logger.error(message)
             return
 
-        try:
-            self._send_remote(url=url, data=data, headers=headers)
-        except Exception, e:
-            if isinstance(e, urllib2.HTTPError):
-                body = e.read()
-                self.error_logger.error('Unable to reach Sentry log server: %s (url: %%s, body: %%s)' % (e,), url, body,
-                    exc_info=True, extra={'data': {'body': body, 'remote_url': url}})
-            else:
-                tmpl = 'Unable to reach Sentry log server: %s (url: %%s)'
-                self.error_logger.error(tmpl % (e,), url, exc_info=True,
-                        extra={'data': {'remote_url': url}})
+        self.logger.debug('Sending message of length %d to %s', len(data), url)
 
-            message = self._get_log_message(data)
-            self.error_logger.error('Failed to submit message: %r', message)
-            self.state.set_fail()
-        else:
-            self.state.set_success()
+        def failed_send(e):
+            self._failed_send(e, url, data)
+
+        try:
+            parsed = urlparse(url)
+            transport = self._registry.get_transport(
+                parsed, **self.transport_options)
+            if transport.async:
+                transport.async_send(data, headers, self._successful_send,
+                                     failed_send)
+            else:
+                transport.send(data, headers)
+                self._successful_send()
+        except Exception as e:
+            failed_send(e)
 
     def send(self, auth_header=None, **data):
         """
@@ -512,13 +594,13 @@ class Client(object):
         """
         Serializes ``data`` into a raw string.
         """
-        return base64.b64encode(json.dumps(data).encode('zlib'))
+        return base64.b64encode(zlib.compress(json.dumps(data).encode('utf8')))
 
     def decode(self, data):
         """
         Unserializes a string, ``data``.
         """
-        return json.loads(base64.b64decode(data).decode('zlib'))
+        return json.loads(zlib.decompress(base64.b64decode(data)).decode('utf8'))
 
     def captureMessage(self, message, **kwargs):
         """
@@ -541,8 +623,11 @@ class Client(object):
         If exc_info is not provided, or is set to True, then this method will
         perform the ``exc_info = sys.exc_info()`` and the requisite clean-up
         for you.
+
+        ``kwargs`` are passed through to ``.capture``.
         """
-        return self.capture('raven.events.Exception', exc_info=exc_info, **kwargs)
+        return self.capture(
+            'raven.events.Exception', exc_info=exc_info, **kwargs)
 
     def captureQuery(self, query, params=(), engine=None, **kwargs):
         """
@@ -550,11 +635,14 @@ class Client(object):
 
         >>> client.captureQuery('SELECT * FROM foo')
         """
-        return self.capture('raven.events.Query', query=query, params=params, engine=engine,
-                **kwargs)
+        return self.capture(
+            'raven.events.Query', query=query, params=params, engine=engine,
+            **kwargs)
 
     def captureExceptions(self, **kwargs):
-        warnings.warn('captureExceptions is deprecated, used context() instead.', DeprecationWarning)
+        warnings.warn(
+            'captureExceptions is deprecated, used context() instead.',
+            DeprecationWarning)
         return self.context(**kwargs)
 
 
