@@ -7,6 +7,7 @@ raven.handlers.logging
 """
 
 from __future__ import absolute_import
+from __future__ import print_function
 
 import datetime
 import logging
@@ -14,8 +15,9 @@ import sys
 import traceback
 
 from raven.base import Client
+from raven.utils import six
 from raven.utils.encoding import to_string
-from raven.utils.stacks import iter_stack_frames
+from raven.utils.stacks import iter_stack_frames, label_from_frame
 
 RESERVED = ('stack', 'name', 'module', 'funcName', 'args', 'msg', 'levelno', 'exc_text', 'exc_info', 'data', 'created', 'levelname', 'msecs', 'relativeCreated', 'tags')
 
@@ -25,7 +27,7 @@ class SentryHandler(logging.Handler, object):
         client = kwargs.get('client_cls', Client)
         if len(args) == 1:
             arg = args[0]
-            if isinstance(arg, basestring):
+            if isinstance(arg, six.string_types):
                 self.client = client(dsn=arg)
             elif isinstance(arg, Client):
                 self.client = arg
@@ -44,67 +46,61 @@ class SentryHandler(logging.Handler, object):
 
         logging.Handler.__init__(self, level=kwargs.get('level', logging.NOTSET))
 
+    def can_record(self, record):
+        return not (
+            record.name == 'raven' or
+            record.name.startswith(('sentry.errors', 'raven.'))
+        )
+
     def emit(self, record):
         try:
+            # Beware to python3 bug (see #10805) if exc_info is (None, None, None)
             self.format(record)
 
-            # Avoid typical config issues by overriding loggers behavior
-            if record.name.startswith('sentry.errors'):
-                print >> sys.stderr, to_string(record.message)
+            if not self.can_record(record):
+                print(to_string(record.message), file=sys.stderr)
                 return
 
             return self._emit(record)
         except Exception:
-            print >> sys.stderr, "Top level Sentry exception caught - failed creating log record"
-            print >> sys.stderr, to_string(record.msg)
-            print >> sys.stderr, to_string(traceback.format_exc())
+            print("Top level Sentry exception caught - failed creating log record", file=sys.stderr)
+            print(to_string(record.msg), file=sys.stderr)
+            print(to_string(traceback.format_exc()), file=sys.stderr)
 
-            try:
-                self.client.captureException()
-            except Exception:
-                pass
+    def _get_targetted_stack(self, stack):
+        # we might need to traverse this multiple times, so coerce it to a list
+        stack = list(stack)
+        frames = []
+        started = False
+        last_mod = ''
+
+        for item in stack:
+            if isinstance(item, (list, tuple)):
+                frame, lineno = item
+            else:
+                frame, lineno = item, item.f_lineno
+
+            if not started:
+                f_globals = getattr(frame, 'f_globals', {})
+                module_name = f_globals.get('__name__', '')
+                if ((last_mod and last_mod.startswith('logging'))
+                        and not module_name.startswith('logging')):
+                    started = True
+                else:
+                    last_mod = module_name
+                    continue
+
+            frames.append((frame, lineno))
+
+        # We failed to find a starting point
+        if not frames:
+            return stack
+
+        return frames
 
     def _emit(self, record, **kwargs):
         data = {}
 
-        for k, v in record.__dict__.iteritems():
-            if '.' not in k and k not in ('culprit',):
-                continue
-            data[k] = v
-
-        stack = getattr(record, 'stack', None)
-        if stack is True:
-            stack = iter_stack_frames()
-
-        if stack:
-            # we might need to traverse this multiple times, so coerce it to a list
-            stack = list(stack)
-            frames = []
-            started = False
-            last_mod = ''
-            for item in stack:
-                if isinstance(item, (list, tuple)):
-                    frame, lineno = item
-                else:
-                    frame, lineno = item, item.f_lineno
-
-                if not started:
-                    f_globals = getattr(frame, 'f_globals', {})
-                    module_name = f_globals.get('__name__', '')
-                    if (last_mod and last_mod.startswith('logging')) \
-                        and not module_name.startswith('logging'):
-                        started = True
-                    else:
-                        last_mod = module_name
-                        continue
-                frames.append((frame, lineno))
-
-            # We must've not found a starting point
-            if not frames:
-                frames = stack
-
-            stack = frames
-
         extra = getattr(record, 'data', None)
         if not isinstance(extra, dict):
             if extra:
@@ -112,19 +108,39 @@ class SentryHandler(logging.Handler, object):
             else:
                 extra = {}
 
-        # Add in all of the data from the record that we aren't already capturing
-        for k in record.__dict__.keys():
+        for k, v in six.iteritems(vars(record)):
             if k in RESERVED:
                 continue
             if k.startswith('_'):
                 continue
-            extra[k] = record.__dict__[k]
+            if '.' not in k and k not in ('culprit',):
+                extra[k] = v
+            else:
+                data[k] = v
+
+        stack = getattr(record, 'stack', None)
+        if stack is True:
+            stack = iter_stack_frames()
+
+        if stack:
+            stack = self._get_targetted_stack(stack)
 
         date = datetime.datetime.utcfromtimestamp(record.created)
         event_type = 'raven.events.Message'
-        handler_kwargs = {'message': record.msg, 'params': record.args}
-        if hasattr(record, 'tags'):
-            handler_kwargs['tags'] = record.tags
+        handler_kwargs = {
+            'params': record.args,
+        }
+        try:
+            handler_kwargs['message'] = six.text_type(record.msg)
+        except UnicodeDecodeError:
+            # Handle binary strings where it should be unicode...
+            handler_kwargs['message'] = repr(record.msg)[1:-1]
+
+        try:
+            handler_kwargs['formatted'] = six.text_type(record.message)
+        except UnicodeDecodeError:
+            # Handle binary strings where it should be unicode...
+            handler_kwargs['formatted'] = repr(record.message)[1:-1]
 
         # If there's no exception being processed, exc_info may be a 3-tuple of None
         # http://docs.python.org/library/sys.html#sys.exc_info
@@ -134,20 +150,24 @@ class SentryHandler(logging.Handler, object):
             # message interface attached
             handler = self.client.get_handler(event_type)
             data.update(handler.capture(**handler_kwargs))
-            # ensure message is propagated, otherwise the exception will overwrite it
-            data['message'] = handler.to_string(data)
 
             event_type = 'raven.events.Exception'
             handler_kwargs = {'exc_info': record.exc_info}
 
         # HACK: discover a culprit when we normally couldn't
         elif not (data.get('sentry.interfaces.Stacktrace') or data.get('culprit')) and (record.name or record.funcName):
-            data['culprit'] = '%s in %s' % (record.name or '<unknown module>', record.funcName or '<unknown function>')
+            culprit = label_from_frame({'module': record.name, 'function': record.funcName})
+            if culprit:
+                data['culprit'] = culprit
 
         data['level'] = record.levelno
         data['logger'] = record.name
 
+        if hasattr(record, 'tags'):
+            kwargs['tags'] = record.tags
+
         kwargs.update(handler_kwargs)
 
-        return self.client.capture(event_type, stack=stack, data=data, extra=extra,
-            date=date, **kwargs)
+        return self.client.capture(
+            event_type, stack=stack, data=data,
+            extra=extra, date=date, **kwargs)
